var apicase = (function () {
'use strict';

function warn () {
  var text = '';
  for (var index in arguments) {
    text += ' ' + arguments[index];
  }
  console.error('[Apicase]' + text);
}

function error () {
  var args = [];
  for (var index in arguments) {
    args.push(arguments[index]);
  }
  warn.apply(null, args.slice(0, -1));
  throw args[args.length - 1]
}

var log = {
  warn,
  error
};

function omit (keys, src) {
  var obj = {};
  for (var key in src) {
    if (!keys.includes(key)) {
      obj[key] = src[key];
    }
  }
  return obj
}

function unary (callback) {
  return function unaryCallback (arg) {
    return callback(arg)
  }
}

var utils = {
  omit,
  unary
};

var isMergeableObject = function isMergeableObject(value) {
	return isNonNullObject(value)
		&& !isSpecial(value)
};

function isNonNullObject(value) {
	return !!value && typeof value === 'object'
}

function isSpecial(value) {
	var stringValue = Object.prototype.toString.call(value);

	return stringValue === '[object RegExp]'
		|| stringValue === '[object Date]'
		|| isReactElement(value)
}

// see https://github.com/facebook/react/blob/b5ac963fb791d1298e7f396236383bc955f916c1/src/isomorphic/classic/element/ReactElement.js#L21-L25
var canUseSymbol = typeof Symbol === 'function' && Symbol.for;
var REACT_ELEMENT_TYPE = canUseSymbol ? Symbol.for('react.element') : 0xeac7;

function isReactElement(value) {
	return value.$$typeof === REACT_ELEMENT_TYPE
}

function emptyTarget(val) {
	return Array.isArray(val) ? [] : {}
}

function cloneUnlessOtherwiseSpecified(value, optionsArgument) {
	var clone = !optionsArgument || optionsArgument.clone !== false;

	return (clone && isMergeableObject(value))
		? deepmerge(emptyTarget(value), value, optionsArgument)
		: value
}

function defaultArrayMerge(target, source, optionsArgument) {
	return target.concat(source).map(function(element) {
		return cloneUnlessOtherwiseSpecified(element, optionsArgument)
	})
}

function mergeObject(target, source, optionsArgument) {
	var destination = {};
	if (isMergeableObject(target)) {
		Object.keys(target).forEach(function(key) {
			destination[key] = cloneUnlessOtherwiseSpecified(target[key], optionsArgument);
		});
	}
	Object.keys(source).forEach(function(key) {
		if (!isMergeableObject(source[key]) || !target[key]) {
			destination[key] = cloneUnlessOtherwiseSpecified(source[key], optionsArgument);
		} else {
			destination[key] = deepmerge(target[key], source[key], optionsArgument);
		}
	});
	return destination
}

function deepmerge(target, source, optionsArgument) {
	var sourceIsArray = Array.isArray(source);
	var targetIsArray = Array.isArray(target);
	var options = optionsArgument || { arrayMerge: defaultArrayMerge };
	var sourceAndTargetTypesMatch = sourceIsArray === targetIsArray;

	if (!sourceAndTargetTypesMatch) {
		return cloneUnlessOtherwiseSpecified(source, optionsArgument)
	} else if (sourceIsArray) {
		var arrayMerge = options.arrayMerge || defaultArrayMerge;
		return arrayMerge(target, source, optionsArgument)
	} else {
		return mergeObject(target, source, optionsArgument)
	}
}

deepmerge.all = function deepmergeAll(array, optionsArgument) {
	if (!Array.isArray(array)) {
		throw new Error('first argument should be an array')
	}

	return array.reduce(function(prev, next) {
		return deepmerge(prev, next, optionsArgument)
	}, {})
};

var deepmerge_1 = deepmerge;




var es = Object.freeze({
	default: deepmerge_1
});

var merge = ( es && deepmerge_1 ) || es;

// TODO: Complete this
function mergeQuery (queries) {
  return merge.apply(null, queries)
}

function reducer (a, b) {
  Object.keys(b).forEach(function (k) {
    a[k] = a[k]
      ? a[k].concat(b[k])
      : [].concat(b[k]);
  });
  return a
}

function mergeWithConcat (options) {
  return options.reduce(reducer, {})
}

function get (object, key, defaultValue) {
  return object[key] || defaultValue
}

var merge_1 = function mergeOptions () {
  var opts = [];
  for (let index in arguments) {
    opts.push(arguments[index]);
  }
  return {
    query: mergeQuery(opts.map(function getQuery (i) {
      return get(i, 'query', {})
    })),
    hooks: mergeWithConcat(opts.map(function getHooks (i) {
      return get(i, 'hooks', {})
    })),
    interceptors: mergeWithConcat(opts.map(function getInterceptors (i) {
      return get(i, 'interceptors', {})
    }))
  }
};

function createAbortCallback (meta) {
  return function abortCallback (reason) {
    meta.isAborted = true;
    meta.abortReason = reason;
  }
}

function createWrapperCreator (type, meta, options) {
  return function createWrapper (cb) {
    return function hookWrapper (ctx, next) {
      try {
        // Calls abortion feature
        if (type === 'before') {
          cb(options, next, createAbortCallback(meta));
        } else {
          cb(ctx, options, next);
        }
        // For future debugging ideas
        if (cb.name && cb.name !== 'endCallback') {
          meta.hooks[type].called++;
        }
      } catch (err) {
        hookErrorLogger(type, cb, meta.hooks[type].called, err);
      }
    }
  }
}

function hookErrorLogger (type, cb, index, err) {
  var name = cb.name && cb.name !== type
    ? cb.name
    : 'anonymous';
  log.warn('Error in ' + type + '[' + index + '] hook');
  log.warn('Hook name: ' + name);
  throw err
}

var hooks = {
  wrapper: createWrapperCreator,
  logger: hookErrorLogger
};

// TODO: add debugging for hook calls
function isAllHooksCalled (type, meta) {}

function isAdapterInstalledCorrectly (name, adapters) {
  var adapter = adapters[name];
  if (
    !adapter ||
    typeof adapter !== 'object' ||
    !adapter.callback ||
    typeof adapter.callback !== 'function'
  ) {
    log.error(
      'Adapter ' + name + ' should be a function or object with callback property',
      '(' + typeof adapter + ' given)',
      new Error('Failed to install adapter')
    );
  }
}

var check = {
  isAllHooksCalled,
  isAdapterInstalledCorrectly
};

var nanoclone = function clone (src) {
  // Null/undefined/functions/etc
  if (
    !src ||
    typeof src !== 'object' ||
    typeof src === 'function'
  ) {
    return src
  }

  // DOM Node
  if (src.nodeType && 'cloneNode' in src) {
    return src.cloneNode(true)
  }

  // Date
  if (src instanceof Date) {
    return new Date(src.getTime())
  }

  // RegExp
  if (src instanceof RegExp) {
    return new RegExp(src)
  }

  // Arrays
  if (Array.isArray(src)) {
    return src.map(clone)
  }

  // Object
  if (src instanceof Object) {
    var obj = {};
    for (var key in src) {
      obj[key] = clone(src[key]);
    }
    return obj
  }

  // ???
  return src
};

/**
 * Expose compositor.
 */

var koaCompose = compose;

/**
 * Compose `middleware` returning
 * a fully valid middleware comprised
 * of all those which are passed.
 *
 * @param {Array} middleware
 * @return {Function}
 * @api public
 */

function compose (middleware) {
  if (!Array.isArray(middleware)) throw new TypeError('Middleware stack must be an array!')
  for (const fn of middleware) {
    if (typeof fn !== 'function') throw new TypeError('Middleware must be composed of functions!')
  }

  /**
   * @param {Object} context
   * @return {Promise}
   * @api public
   */

  return function (context, next) {
    // last called middleware #
    let index = -1;
    return dispatch(0)
    function dispatch (i) {
      if (i <= index) return Promise.reject(new Error('next() called multiple times'))
      index = i;
      let fn = middleware[i];
      if (i === middleware.length) fn = next;
      if (!fn) return Promise.resolve()
      try {
        return Promise.resolve(fn(context, function next () {
          return dispatch(i + 1)
        }))
      } catch (err) {
        return Promise.reject(err)
      }
    }
  }
}

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

var nanoevents = createCommonjsModule(function (module) {
(
  /**
   * Interface for event subscription.
   *
   * @example
   * var NanoEvents = require('nanoevents')
   *
   * class Ticker {
   *   constructor() {
   *     this.emitter = new NanoEvents()
   *   }
   *   on() {
   *     return this.emitter.on.apply(this.events, arguments)
   *   }
   *   tick() {
   *     this.emitter.emit('tick')
   *   }
   * }
   *
   * @alias NanoEvents
   * @class
   */
  module.exports = function NanoEvents () {
    /**
     * Event names in keys and arrays with listeners in values.
     * @type {object}
     *
     * @example
     * Object.keys(ee.events)
     *
     * @alias NanoEvents#events
     */
    this.events = { };
  }
).prototype = {

  /**
   * Add a listener for a given event.
   * @param {string} event The event name.
   * @param {function} cb The listener function.
   *
   * @return {function} Unbind listener from event.
   *
   * @example
   * const unbind = ee.on('tick', (tickType, tickDuration) => {
   *   count += 1
   * })
   *
   * disable () {
   *   unbind()
   * }
   *
   * @alias NanoEvents#on
   * @method
   */
  on: function on (event, cb) {
    event = this.events[event] = this.events[event] || [];
    event.push(cb);

    return function () {
      // a.splice(i >>> 0, 1) === if (i !== -1) a.splice(i, 1)
      // -1 >>> 0 === 0xFFFFFFFF, max possible array length
      event.splice(event.indexOf(cb) >>> 0, 1);
    }
  },

  /**
   * Calls each of the listeners registered for a given event.
   *
   * @param {string} event The event name.
   * @param {...*} arguments The arguments for listeners.
   *
   * @returns {undefined}
   *
   * @example
   * ee.emit('tick', tickType, tickDuration)
   *
   * @alias NanoEvents#emit
   * @method
   */
  emit: function emit (event) {
    var list = this.events[event];
    if (!list || !list[0]) return // list[0] === Array.isArray(list)

    var args = list.slice.call(arguments, 1);
    list.slice().map(function (i) {
      i.apply(this, args); // this === global or window
    });
  }
};
});

/**
 * Expose `pathToRegexp`.
 */
var pathToRegexp_1 = pathToRegexp;
var parse_1 = parse;
var compile_1 = compile;
var tokensToFunction_1 = tokensToFunction;
var tokensToRegExp_1 = tokensToRegExp;

/**
 * Default configs.
 */
var DEFAULT_DELIMITER = '/';
var DEFAULT_DELIMITERS = './';

/**
 * The main path matching regexp utility.
 *
 * @type {RegExp}
 */
var PATH_REGEXP = new RegExp([
  // Match escaped characters that would otherwise appear in future matches.
  // This allows the user to escape special characters that won't transform.
  '(\\\\.)',
  // Match Express-style parameters and un-named parameters with a prefix
  // and optional suffixes. Matches appear as:
  //
  // "/:test(\\d+)?" => ["/", "test", "\d+", undefined, "?"]
  // "/route(\\d+)"  => [undefined, undefined, undefined, "\d+", undefined]
  '(?:\\:(\\w+)(?:\\(((?:\\\\.|[^\\\\()])+)\\))?|\\(((?:\\\\.|[^\\\\()])+)\\))([+*?])?'
].join('|'), 'g');

/**
 * Parse a string for the raw tokens.
 *
 * @param  {string}  str
 * @param  {Object=} options
 * @return {!Array}
 */
function parse (str, options) {
  var tokens = [];
  var key = 0;
  var index = 0;
  var path = '';
  var defaultDelimiter = (options && options.delimiter) || DEFAULT_DELIMITER;
  var delimiters = (options && options.delimiters) || DEFAULT_DELIMITERS;
  var pathEscaped = false;
  var res;

  while ((res = PATH_REGEXP.exec(str)) !== null) {
    var m = res[0];
    var escaped = res[1];
    var offset = res.index;
    path += str.slice(index, offset);
    index = offset + m.length;

    // Ignore already escaped sequences.
    if (escaped) {
      path += escaped[1];
      pathEscaped = true;
      continue
    }

    var prev = '';
    var next = str[index];
    var name = res[2];
    var capture = res[3];
    var group = res[4];
    var modifier = res[5];

    if (!pathEscaped && path.length) {
      var k = path.length - 1;

      if (delimiters.indexOf(path[k]) > -1) {
        prev = path[k];
        path = path.slice(0, k);
      }
    }

    // Push the current path onto the tokens.
    if (path) {
      tokens.push(path);
      path = '';
      pathEscaped = false;
    }

    var partial = prev !== '' && next !== undefined && next !== prev;
    var repeat = modifier === '+' || modifier === '*';
    var optional = modifier === '?' || modifier === '*';
    var delimiter = prev || defaultDelimiter;
    var pattern = capture || group;

    tokens.push({
      name: name || key++,
      prefix: prev,
      delimiter: delimiter,
      optional: optional,
      repeat: repeat,
      partial: partial,
      pattern: pattern ? escapeGroup(pattern) : '[^' + escapeString(delimiter) + ']+?'
    });
  }

  // Push any remaining characters.
  if (path || index < str.length) {
    tokens.push(path + str.substr(index));
  }

  return tokens
}

/**
 * Compile a string to a template function for the path.
 *
 * @param  {string}             str
 * @param  {Object=}            options
 * @return {!function(Object=, Object=)}
 */
function compile (str, options) {
  return tokensToFunction(parse(str, options))
}

/**
 * Expose a method for transforming tokens into the path function.
 */
function tokensToFunction (tokens) {
  // Compile all the tokens into regexps.
  var matches = new Array(tokens.length);

  // Compile all the patterns before compilation.
  for (var i = 0; i < tokens.length; i++) {
    if (typeof tokens[i] === 'object') {
      matches[i] = new RegExp('^(?:' + tokens[i].pattern + ')$');
    }
  }

  return function (data, options) {
    var path = '';
    var encode = (options && options.encode) || encodeURIComponent;

    for (var i = 0; i < tokens.length; i++) {
      var token = tokens[i];

      if (typeof token === 'string') {
        path += token;
        continue
      }

      var value = data ? data[token.name] : undefined;
      var segment;

      if (Array.isArray(value)) {
        if (!token.repeat) {
          throw new TypeError('Expected "' + token.name + '" to not repeat, but got array')
        }

        if (value.length === 0) {
          if (token.optional) continue

          throw new TypeError('Expected "' + token.name + '" to not be empty')
        }

        for (var j = 0; j < value.length; j++) {
          segment = encode(value[j]);

          if (!matches[i].test(segment)) {
            throw new TypeError('Expected all "' + token.name + '" to match "' + token.pattern + '"')
          }

          path += (j === 0 ? token.prefix : token.delimiter) + segment;
        }

        continue
      }

      if (typeof value === 'string' || typeof value === 'number' || typeof value === 'boolean') {
        segment = encode(String(value));

        if (!matches[i].test(segment)) {
          throw new TypeError('Expected "' + token.name + '" to match "' + token.pattern + '", but got "' + segment + '"')
        }

        path += token.prefix + segment;
        continue
      }

      if (token.optional) {
        // Prepend partial segment prefixes.
        if (token.partial) path += token.prefix;

        continue
      }

      throw new TypeError('Expected "' + token.name + '" to be ' + (token.repeat ? 'an array' : 'a string'))
    }

    return path
  }
}

/**
 * Escape a regular expression string.
 *
 * @param  {string} str
 * @return {string}
 */
function escapeString (str) {
  return str.replace(/([.+*?=^!:${}()[\]|/\\])/g, '\\$1')
}

/**
 * Escape the capturing group by escaping special characters and meaning.
 *
 * @param  {string} group
 * @return {string}
 */
function escapeGroup (group) {
  return group.replace(/([=!:$/()])/g, '\\$1')
}

/**
 * Get the flags for a regexp from the options.
 *
 * @param  {Object} options
 * @return {string}
 */
function flags (options) {
  return options && options.sensitive ? '' : 'i'
}

/**
 * Pull out keys from a regexp.
 *
 * @param  {!RegExp} path
 * @param  {Array=}  keys
 * @return {!RegExp}
 */
function regexpToRegexp (path, keys) {
  if (!keys) return path

  // Use a negative lookahead to match only capturing groups.
  var groups = path.source.match(/\((?!\?)/g);

  if (groups) {
    for (var i = 0; i < groups.length; i++) {
      keys.push({
        name: i,
        prefix: null,
        delimiter: null,
        optional: false,
        repeat: false,
        partial: false,
        pattern: null
      });
    }
  }

  return path
}

/**
 * Transform an array into a regexp.
 *
 * @param  {!Array}  path
 * @param  {Array=}  keys
 * @param  {Object=} options
 * @return {!RegExp}
 */
function arrayToRegexp (path, keys, options) {
  var parts = [];

  for (var i = 0; i < path.length; i++) {
    parts.push(pathToRegexp(path[i], keys, options).source);
  }

  return new RegExp('(?:' + parts.join('|') + ')', flags(options))
}

/**
 * Create a path regexp from string input.
 *
 * @param  {string}  path
 * @param  {Array=}  keys
 * @param  {Object=} options
 * @return {!RegExp}
 */
function stringToRegexp (path, keys, options) {
  return tokensToRegExp(parse(path, options), keys, options)
}

/**
 * Expose a function for taking tokens and returning a RegExp.
 *
 * @param  {!Array}  tokens
 * @param  {Array=}  keys
 * @param  {Object=} options
 * @return {!RegExp}
 */
function tokensToRegExp (tokens, keys, options) {
  options = options || {};

  var strict = options.strict;
  var end = options.end !== false;
  var delimiter = escapeString(options.delimiter || DEFAULT_DELIMITER);
  var delimiters = options.delimiters || DEFAULT_DELIMITERS;
  var endsWith = [].concat(options.endsWith || []).map(escapeString).concat('$').join('|');
  var route = '';
  var isEndDelimited = false;

  // Iterate over the tokens and create our regexp string.
  for (var i = 0; i < tokens.length; i++) {
    var token = tokens[i];

    if (typeof token === 'string') {
      route += escapeString(token);
      isEndDelimited = i === tokens.length - 1 && delimiters.indexOf(token[token.length - 1]) > -1;
    } else {
      var prefix = escapeString(token.prefix);
      var capture = token.repeat
        ? '(?:' + token.pattern + ')(?:' + prefix + '(?:' + token.pattern + '))*'
        : token.pattern;

      if (keys) keys.push(token);

      if (token.optional) {
        if (token.partial) {
          route += prefix + '(' + capture + ')?';
        } else {
          route += '(?:' + prefix + '(' + capture + '))?';
        }
      } else {
        route += prefix + '(' + capture + ')';
      }
    }
  }

  if (end) {
    if (!strict) route += '(?:' + delimiter + ')?';

    route += endsWith === '$' ? '$' : '(?=' + endsWith + ')';
  } else {
    if (!strict) route += '(?:' + delimiter + '(?=' + endsWith + '))?';
    if (!isEndDelimited) route += '(?=' + delimiter + '|' + endsWith + ')';
  }

  return new RegExp('^' + route, flags(options))
}

/**
 * Normalize the given path string, returning a regular expression.
 *
 * An empty array can be passed in for the keys, which will hold the
 * placeholder key descriptions. For example, using `/user/:id`, `keys` will
 * contain `[{ name: 'id', delimiter: '/', optional: false, repeat: false }]`.
 *
 * @param  {(string|RegExp|Array)} path
 * @param  {Array=}                keys
 * @param  {Object=}               options
 * @return {!RegExp}
 */
function pathToRegexp (path, keys, options) {
  if (path instanceof RegExp) {
    return regexpToRegexp(path, keys)
  }

  if (Array.isArray(path)) {
    return arrayToRegexp(/** @type {!Array} */ (path), keys, options)
  }

  return stringToRegexp(/** @type {string} */ (path), keys, options)
}

pathToRegexp_1.parse = parse_1;
pathToRegexp_1.compile = compile_1;
pathToRegexp_1.tokensToFunction = tokensToFunction_1;
pathToRegexp_1.tokensToRegExp = tokensToRegExp_1;

function compilePath (url, params) {
  return pathToRegexp_1.compile(url)(params)
}

function evaluateHeaders (headers) {
  return typeof headers === 'function'
    ? headers()
    : headers
}

function buildQueryString (query) {
  var queryString = '';
  Object.keys(query).forEach(function encodeQueryPart (key) {
    queryString += encodeURIComponent(key) + '=' + encodeURIComponent(query[key]);
  });
  return queryString.length
    ? '?' + queryString
    : ''
}

function defaultValidator (target, event) {
  return (target.status >= 200 && target.status <= 299) || target.status === 304
}

function xhrAdapter (ctx) {
  var xhr = new XMLHttpRequest();
  var url = compilePath(ctx.options.url, ctx.options.params || {});
  var options = {
    method: ctx.options.method || 'GET',
    body: ctx.options.body || null,
    headers: ctx.options.headers ? evaluateHeaders(ctx.options.headers) : {}
  };
  if (ctx.options.query) url += buildQueryString(ctx.options.query);
  var isSuccess = ctx.options.validator || defaultValidator;

  xhr.open(options.method, url, true);
  xhr.onprogress = function onProgress (e) {
    ctx.customHook('progress', e);
  };
  xhr.onload = function onLoad (e) {
    if (isSuccess(e.target, e)) {
      ctx.done(e);
    } else {
      ctx.fail(e);
    }
  };
  xhr.onabort = function onAbort (e) {
    ctx.customHook('aborted', e, true);
  };
  for (var key in options.headers) {
    xhr.setRequestHeader(key, options.headers[key]);
  }
  xhr.send(options.body);
}

var apicaseAdapterXhr = xhrAdapter;

function compilePath$1 (url, params) {
  return pathToRegexp_1.compile(url)(params)
}

function evaluateHeaders$1 (headers) {
  return typeof headers === 'function'
    ? headers()
    : headers
}

function buildQueryString$1 (query) {
  var queryString = '';
  Object.keys(query).forEach(function encodeQueryPart (key) {
    queryString += encodeURIComponent(key) + '=' + encodeURIComponent(query[key]);
  });
  return queryString.length
    ? '?' + queryString
    : ''
}

function fetchAdapter (ctx) {
  var options = {
    method: ctx.options.method || 'GET',
    credentials: ctx.options.credentials || 'omit'
  };
  var url = compilePath$1(ctx.options.url, ctx.options.params || {});
  if (ctx.options.body) options.body = ctx.options.body;
  if (ctx.options.headers) options.headers = evaluateHeaders$1(ctx.options.headers);
  if (ctx.options.query) url += buildQueryString$1(ctx.options.query);

  return fetch(url, options)
    .then(function resolveFetch (res) {
      if (res.ok) {
        return res[ctx.options.parser || 'json']()
      } else {
        throw res
      }
    })
    .then(function resolveAdapter (data) {
      ctx.done(data);
    })
    .catch(function rejectAdapter (reason) {
      ctx.fail(reason);
    })
}

var apicaseAdapterFetch = fetchAdapter;

// Global config is outside Apicase
// because of no need to store, mutate and clone it there
var globalConfig = {
  silent: false,
  defaultAdapter: 'fetch',
  adapters: {
    xhr: { callback: apicaseAdapterXhr },
    fetch: { callback: apicaseAdapterFetch }
  }
};

var Apicase = function (options) {
  this.base = {
    query: {},
    hooks: {
      before: [],
      success: [],
      error: [],
      aborted: []
    },
    interceptors: {
      before: [],
      success: [],
      error: [],
      aborted: []
    }
  };

  // Call installer with apicase instance and additional options
  // use it in plugins to add new features
  this.use = function (installer, options) {
    this.bus.emit('preinstall', this);
    installer(this, options);
    this.bus.emit('postinstall', this);
  };

  // Like .use but clones instance
  this.extend = function (installer, options) {
    var cloned = nanoclone(this);
    cloned.use(installer, options);
    Object.setPrototypeOf(cloned, this);
    return cloned
  };

  // Creates a service from optiosn
  // you can pass a part of query/hooks/interceptors
  // to reduce boilerplate code
  this.service = function (options) {
    var instance = this;
    var cloned = Object.assign({}, instance, {
      base: merge_1(instance.base, {
        hooks: options.hooks,
        query: utils.omit(['hooks', 'interceptors'], options),
        interceptors: options.interceptors
      }),
      bus: nanoclone(instance.bus)
    });
    Object.setPrototypeOf(cloned, this);
    return cloned
  };

  // Make API request
  this.call = function () {
    var instance = this;

    // Apicase.call(1, 2, 3) now is equal with
    // Promise.all([Apicase.call(1), Apicase.call(2), Apicase.call(3)])
    if (arguments.length > 1) {
      var queries = Array.prototype.slice.call(arguments);
      return Promise.all(queries.map(utils.unary(instance.call)))
    }

    // First argument is options object
    var options = arguments[0];

    // Meta info about hook calls (for future debugging needs)
    // + calls abortion
    var meta = {
      isAborted: false,
      abortReason: null,
      hooks: {}
    };

    // Merged options
    // TODO: Merge strategies for query
    var o = merge_1(this.base, !options ? {} : {
      hooks: options.hooks,
      query: utils.omit(['hooks', 'interceptors'], options),
      interceptors: options.interceptors
    });

    // Adapter callback
    var adapterName = o.query.adapter || globalConfig.defaultAdapter;
    var adapter = globalConfig.adapters[adapterName];
    if (!adapter) {
      log.error(
        'Adapter ' + adapterName + ' not found',
        new ReferenceError('Adapter ' + adapterName + ' not found')
      );
    }

    // Call interceptors of needed type for passed contenxt
    function callInterceptors (type, context, cbMeta) {
      var result = {
        isStateChanged: false,
        context: nanoclone(context)
      };
      var interceptors = o.interceptors[type];
      function changeState (nextContext) {
        result.isStateChanged = true;
        result.context = nextContext;
      }
      if (!interceptors) return result
      for (var index in interceptors) {
        var stepContext =
          type === 'before'
            ? interceptors[index](context)
            : interceptors[index](context, changeState, cbMeta || { isStateChanged: false });
        if (result.isStateChanged) return result
        result.context = stepContext;
      }

      return result
    }

    // Call hooks of needed type with passed context
    // also accepts additional callbacks
    function callHooks (type, context, additionalCb) {
      // I really don't want to mutate context/options in hooks
      // use interceptors instead
      var clonedContext = nanoclone(context);
      var clonedOptions = nanoclone(o.query);

      // Meta info for debugging
      meta.hooks[type] = {
        called: 0,
        all: (o.hooks[type] && o.hooks[type].length) || 0
      };

      var h = (o.hooks[type] || [])
        .concat(additionalCb || [])
        .map(hooks.wrapper(type, meta, clonedOptions));

      instance.emit(type, clonedContext, clonedOptions);

      return koaCompose(h)(clonedContext)
        .then(function checkHookCalls () {
          if (!globalConfig.silent) {
            check.isAllHooksCalled(type, meta);
          }
        })
    }

    // Call adapter and then do some things
    // TODO: reduce boilerplate code
    function callAdapter (resolve, reject) {
      var done = function doneCallback (data, cbMeta) {
        var res = callInterceptors('success', data, cbMeta);

        if (res.isStateChanged) {
          fail(res.context, { isStateChanged: true });
        } else {
          callHooks('success', res.context, resolve)
            .then(() => {
              callHooks('finished', res.context);
            });
        }
      };

      var fail = function failCallback (reason, cbMeta) {
        var res = callInterceptors('error', reason, cbMeta);

        if (res.isStateChanged) {
          done(res.context, { isStateChanged: true });
        } else {
          callHooks('error', res.context, reject)
            .then(() => {
              callHooks('finished', res.context);
            });
        }
      };

      var customHook = function customCallback (type, data, isError, cbMeta) {
        var res = callInterceptors(type, data, cbMeta);

        // If custom hook should reject promise
        // but interceptors are changed state
        // it will be resolved
        // otherwise it will be rejected
        if (res.isStateChanged) {
          if (isError) {
            done(res.context, { isStateChanged: true });
          } else {
            fail(res.context, { isStateChanged: true });
          }
        } else {
          callHooks(type, res.context, isError ? reject : null);
        }
      };

      var callAdapter = function callAnotherAdapter (name, options) {
        if (!(name in instance.options.adapters)) {
          log.error(
            'Adapter ' + name + ' not found',
            'in high-order adapter' + o.query.adapter,
            new ReferenceError('Adapter ' + name + ' not found')
          );
        }
        var adapter = instance.options.adapters[name];
        return adapter.callback({
          adapter,
          instance,
          done,
          fail,
          customHook,
          callAdapter,
          options: options === undefined
            ? o.query
            : options
        })
      };

      adapter.callback({
        adapter,
        instance,
        done,
        fail,
        customHook,
        callAdapter,
        options: o.query
      });
    }
    // Transform query with interceptors
    o.query = callInterceptors('before', o.query).context;

    return new Promise(function makeCall (resolve, reject) {
      callHooks('before', {})
        .then(function onBeforeHooksSuccess () {
          if (meta.isAborted) {
            callHooks('aborted', { reason: meta.abortReason }, reject);
          } else {
            callAdapter(resolve, reject);
          }
        });
    })
  };

  // CAUTION: It will be removed in next version
  // Now it just calls apicase.call(...queries)
  this.all = function (options) {
    return this.call.apply(this, options)
  };

  // Returns instance with .call() wrapped into a function
  // that make calls with options converted by a callback
  this.transform = function (wrapper) {
    return this.extend(function callTransformer (i) {
      var realCall = nanoclone(i.call).bind(i);

      i.call = function transformer () {
        var options = Array.prototype.slice.call(arguments)
          .map(utils.unary(wrapper));

        return options.length <= 1
          ? realCall(options[0])
          : Promise.all(options.map(utils.unary(realCall)))
      };
    })
  };

  // Like .transform() but does .call(...options) for returned array
  this.transformSpread = function (wrapper) {
    return this.extend(function callSpreadTransformer (i) {
      var realCall = nanoclone(i.call).bind(i);

      i.call = function spreadTransformer () {
        var options = Array.prototype.slice.call(arguments)
          .reduce((acc, cur) => acc.concat(wrapper(cur)), []);

        return options.length <= 1
          ? realCall(options[0])
          : Promise.all(options.map(utils.unary(realCall)))
      };
    })
  };

  // Calls chaining with any options
  // accepts objects / functions with passed previous result
  // or apicase instances (will use .call() method)
  this.queue = function (options, lastData) {
    if (!options.length) return lastData
    var instance = this;

    // Just call queue again without first data
    function nextQueue (data) {
      return instance.queue(options.slice(1), data)
    }

    // Use .call() if it's Apicase instance
    if (options[0] instanceof Apicase) {
      return options[0].call(lastData)
        .then(nextQueue)
    }

    // Make a call with options object
    // or with options returned by function passed
    var option =
      typeof options[0] === 'function'
        ? options[0](lastData)
        : options[0];

    return instance
      .call(option)
      .then(nextQueue)
  };

  // Events bus to handle service calls anywhere
  this.bus = new nanoevents();

  // Shortcut for .bus.on(event, callback)
  this.on = function (event, callback) {
    return this.bus.on(event, callback)
  };

  // Shortcut for .bus.emit(event, ...arguments)
  this.emit = function () {
    this.bus.emit.apply(this.bus, arguments);
  };

  return this
};

// Add new adapter
Apicase.addAdapter = function (name, adapter) {
  globalConfig.adapters[name] =
    typeof adapter === 'function'
      ? { callback: adapter }
      : adapter;
  if (!globalConfig.silent) {
    check.isAdapterInstalledCorrectly(name, globalConfig.adapters);
  }
};

// Add silent property from global config to Apicase
Object.defineProperty(Apicase, 'silent', {
  get () {
    return globalConfig.silent
  },
  set (val) {
    globalConfig.silent = val;
  }
});

var apicaseCore = Apicase;

return apicaseCore;

}());
